# -*- coding: utf-8 -*-
#normaliser for XHTML input
"""
Takes (possibly bad) html chunks and normalises to a valid XHTML subset.

Borrows from BeautifulSoup, and StrippingParser in ActiveState cookbook.

Previous attempts at a normalize function have gone badly as
we ended up with a huge mix of approaches:  search-and-replaces,
Soup, Tidy, regexes and so on.  We end up with them fighting
each other.

Also, we failed to define our intended context clearly enough.
When you want to use someone else's content in RML, typically
you know whether you're using it like this...

<h2>{{attraction.name}}<h2>

..or like this...

<h2>Some heading</h2>
{{possibleMultipleParagraphsHere}}


The HTML Cleaner asks you to specify the target context, and then cleans
and normalizes your code to match.  The full language is as follows:


Block level tags:  p, h1, h2, h3, h4, h5, table, ul, ol, table

Inline tags: i, b, u, br, a, img

Others:

  table contains tr/td
  ul/ol contains li   (are nested lists OK?)
  
Plain text can only appear inside blocks, never between them.



Requirements
============
Accepts 8-bit content.  Assume it's UTF8, but if we get
illegal characters, try a conversion from Latin-1.

Fix naked ampersands and unescaped < or >.  Japanese sometimes
does "<<phrase>>" for emphasis.

Fix case problems in tags (<I> etc).

Close unclosed tags where possible, raise exception for gross
imbalances.

Gather warnings for inspection.

Strip out all but a small, defined set of tags and attributes




"""
import re
import sgmllib
from copy import deepcopy
from xml.sax.saxutils import escape, quoteattr
from rlextra.utils.tagstripper import stripTags

#this is what we look for with doubled entities... e.g. &amp;amp;
leftover_entity = re.compile('^(([a-zA-Z][-.a-zA-Z0-9]*)|(#[0-9]+));')



def cleanBlocks(input, **options):
    """Accept markup as one or more blocks.

    The output of this should be safe for use within a 
    <div> or <body> tag in HTML, and also convertible to RML.
    
    """
    return Cleaner(target='block', **options).process(input)

def cleanInline(input, **options):
    """Accept and normalize markup for use inline.
    
    The output of this should be safe for use within a 
    <p> tag in HTML, and also convertible to RML.
    """
    return Cleaner(target='inline', **options).process(input)

def cleanPlain(input, **options):
    """Remove all tags to output plain text.
    
    """
    return stripTags(input)


class Cleaner(sgmllib.SGMLParser):

    from htmlentitydefs import entitydefs # replace entitydefs from sgmllib

    def __init__(self, target="block", breaksAllowed=True, convertEntities=False, stripComments=False):
        """Initialising defines your language options.
        You can re-use the same parser many times.
        if breaksAllowed, they will be written to output.
        if not, in inline mode they vanish, and in block
        mode they end the block.
        """
        self._setupGrammar()
        assert target in ("block", "inline"), "unexpected block '%s', must be 'block' or 'inline'" % mode
        self.target = target
        sgmllib.SGMLParser.__init__(self)
        self.requireWellFormed = False    #must tags match?
        self.breaksAllowed = breaksAllowed
        self.convertEntities = convertEntities
        self.stripComments = stripComments

    def _setupGrammar(self):

        self.DEFAULT_BLOCK_TAG = 'p'   #used to close blocks if nothing else given


        def byName(listOfDicts):
            out = {}
            for d in listOfDicts:
               out[d['name']] = d
            return out

        #situations in which a tag can appear.
        self._contextDict = byName([
            dict(name='block', allowedIn=None),  #None=top-level
            dict(name='inline', allowedIn='block'),
            dict(name='table', allowedIn='block'),
            dict(name='list', allowedIn='block'),
            dict(name='li', allowedIn='list'),
            dict(name='tr', allowedIn='table'),
            dict(name='td', allowedIn='tr'),  #no contents for now
            ])

        for context in self._contextDict.values():
            context['canContain'] = set()


        #tags themselves
        self._tagDict = byName([
            dict(name='p',context='block', attrs=['style','align']),
            dict(name='h1',context='block', attrs=['style','align']),
            dict(name='h2',context='block', attrs=['style','align']),
            dict(name='h3',context='block', attrs=['style','align']),
            dict(name='h4',context='block', attrs=['style','align']),
            dict(name='h5',context='block', attrs=['style','align']),
            dict(name='h5',context='block', attrs=['style','align']),

            dict(name='i',context='inline', attrs=[]),
            dict(name='b',context='inline', attrs=[]),
            dict(name='u',context='inline', attrs=[]),
            dict(name='br',context='inline', attrs=[], selfClosing=True),
            dict(name='a',context='inline', attrs=['href']),

            # force_attrs. These attributes will be added to make sure xhtml validation passes.
            dict(name='img',context='inline', attrs=['src','width','height','alt'], force_attrs=['alt'], selfClosing=True),

            dict(name='table',context='block', attrs=[]),
            dict(name='tr',context='table', attrs=[]),
            dict(name='td',context='tr', attrs=[]),

            dict(name='ul',context='block', attrs=[]),
            dict(name='ol',context='block', attrs=[]),
            dict(name='li',context='list', attrs=[]),
            ])


        #work out in-place the set of tags allowed in each context.

        for tagName in self._tagDict:
            tag = self._tagDict[tagName]
            if not tag.has_key('selfClosing'):
                tag['selfClosing'] = False
                
            ctxName = tag['context']

            context = self._contextDict.get(ctxName)
            context['canContain'].add(tagName)



        #work out certain useful attributes
        self.valid_tags = set(self._tagDict.keys())
        self.valid_block_tags = self.tagsAllowedInContext('block')
        self.valid_inline_tags = self.tagsAllowedInContext('inline')
        self.valid_other_tags = self.valid_tags - self.valid_block_tags - self.valid_inline_tags



    def allowedAttrs(self, tagName):
        """Return set of allowed attributes for the tag"""
        return self._tagDict[tagName]['attrs']

    def forcedAttrs(self, tagName):
        """Return set of forced attributes for the tag"""
        if 'force_attrs' in self._tagDict[tagName]:
            return self._tagDict[tagName]['force_attrs']
        else:
            return None
    
    def getContext(self, tag):
        """Return context for tag

        >>> g = Cleaner()
        >>> g.getContext('i')
        'inline'
        >>> g.getContext('li')
        'list'
        """
        return self._tagDict[tag]['context']

    def isTagAllowedInContext(self, tag, context):
        """Is the tag allowed here?

        >>> g = Cleaner()
        >>> g.isTagAllowedInContext('b','block')
        False
        >>> g.isTagAllowedInContext('a','inline')
        True


        """
        return (self._tagDict[tag]['context'] == context)

    def tagsAllowedInContext(self, context):
        """Set of tag names allowed in this context

        >>> g = Cleaner()
        >>> g.tagsAllowedInContext('table')
        set(['tr'])
        >>> g.tagsAllowedInContext('inline')
        set(['a', 'i', 'b', 'u', 'br'])
        """
        return self._contextDict[context]['canContain']


    def reset(self):
        "get ready to do some work"
        sgmllib.SGMLParser.reset(self)
        if self.target == 'block':
            self._context = 'block'    #expect to see block tags
        else:
            self._context = 'inline'   #expect to see inline tags
        self.buf = []   #holds output
        self.fixes = []  #holds warnings / debug messages / fixups done
        self.openTagStack = []  #checks for balancing
        self._started = False
        self._currentBlockTag = None   #what kind of block tag are we inside?  Usually <p>
        self._justAfterEntity = False   #flag to say if the last thing we saw was an entity.  Used to detect doubled entities in input
    def close(self):
        "Final tidyups"
        sgmllib.SGMLParser.close(self)
        self.cleanupClosingTags()



    def process(self, markup):
        "The main loop - call this with your markup"
        markup = markup.strip()
        # <br/> needs to be converted into <br />
        markup = markup.replace('<br/>', '<br />')
        self.reset()
        markup = self.fixEncoding(markup)
        self.feed(markup)
        self.close()
        return ''.join(self.buf)


    def dump(self):
        print ''.join(self.buf)

    def fixEncoding(self, markup):
        """To be completed.

        At minimum, check it's valid utf8.  If possible and if it
        is not, detect encoding or assume cp1252.  Look at
        UnicodeDammit in BeautifulSoup.
        """
        if type(markup) == type(u''):
            out = markup.encode('utf8')
            print 'converted unicode to utf8'
        else:
            try:
                uni = markup.decode('utf8', 'strict')
                #OK, it was valid, markup not modified
                out = markup
            except UnicodeDecodeError, e:
                #assume windows encoding
                uni = markup.decode('cp1252', 'replace')
                out = uni.encode('utf8')
        return out

    def writeStartTag(self, tag, attrs={}):
        """Helper to do what it says.  Called to write a tag to output.

        Never write your own tags to output; instead call this.  This will
        maintain a stack and ensure they are balanced. It also sets the
        mode every time for you."""

        #self.dump()
        #prefilter to remove all block tags in inline markup

        if tag not in self.valid_inline_tags:
            if self.target == 'inline':
                return

        allowedAttrs = self.allowedAttrs(tag)
        forcedAttrs = self.forcedAttrs(tag)
        selfClosing = self._tagDict[tag]['selfClosing']
        #if selfClosing: print "found self-closing tag %s" % tag

        #rebuild the tag as a piece of text
        tagBits = ['<']
        tagBits.append(tag)
        for k, v in attrs:
            v = self.fixEncoding(v)
            if k in allowedAttrs:
                if k[0:2].lower() != 'on' and v[0:10].lower() != 'javascript':
                    tagBits.append(' %s="%s"' % (k, v))

        # If there are any forced attributes
        if forcedAttrs and len(forcedAttrs) > 0:
            tag_attrs = [k for k,v in attrs]
            for k in forcedAttrs:
                if k not in tag_attrs:
                    blank = ''
                    tagBits.append(' %s="%s"'%(k, blank))
        if selfClosing:
            tagBits.append('/>')
        else:
            tagBits.append('>')
        tagText = ''.join(tagBits)

        self.buf.append(tagText)

        #and put it on the stack....
        if not selfClosing:
            self.openTagStack.append(tag)
            

            #if block, remember how to close
            if self._context == 'block':
                self._currentBlockTag = tag
    
            #set the mode
            if tag == 'table':
                self._context = 'table'
            elif tag == 'tr':
                self._context = 'tr'
            elif tag == 'td':
                self._context = 'td'
            elif tag in ('ul', 'ol'):
                self._context = 'list'
            elif tag == 'li':
                self._context = 'li'
            else:
                #block and inline always lead to inline
                self._context = 'inline'
        else:
            self._context = self.getContext(tag)
            #print 
            #print "wrote an img tag.  context after = %s" % self._context
            #print "output after: " + ''.join(self.buf)
            #print "stack after: " + str(self.openTagStack)
            


    def writeEndTag(self, tag):
        """Close the tag, but check for nesting errors.

        Never write to the buffer directly; this keeps the stack
        and mode organised."""

        #self.dump()
        #prefilter to remove all block tags in inline markup
        if self.target == 'inline':
            if tag not in self.valid_inline_tags:
                return
        lastTag = self.openTagStack.pop()
        if lastTag != tag:
            raise ValueError("stack is messed up trying to close %s; current open tag was %s" % (tag, lastTag))
        self.buf.append('</%s>' % tag)

        self._context = self.getContext(tag)


    def pendingTag(self):
        "What tag is waiting?"
        try:
            return self.openTagStack[-1]
        except IndexError:
            return None

    def atStart(self):
        return (len(self.buf) == 0)

    def writeData(self, text):
        "Used to write out non-tag content"
        self.buf.append(text)

    def closeCurrentBlock(self):
        """This is used to close any pending inline tags, whenever
        we hit a new block start,  Healthy closing never calls this."""
        if self.atStart():
            return
        if self.target == 'inline':
            return #write nothing
        tag = self.pendingTag()
        if tag is not None:
            while tag in self.valid_inline_tags:
                self.writeEndTag(tag)
                tag = self.pendingTag()
        assert self._currentBlockTag is not None
            
        self.buf.append("</%s>" % self._currentBlockTag)
        self._context = 'block'

        #if there are any more end-tags in the stack, chuck them.
        self.openTagStack = []



    def handle_data(self, data):
        #apply our text cleanups here....the parser will already have
        #recognised any genuine entities, so we can safely escape what's
        #left to deal with naked ampersands.  The escape function just
        #does the 3 standard search-and-replaces for &, < and >.
    
        #it might be the second half of an &amp;amp; mistake in the input.
        #Since sgmllib has already process the first &amp;, we have to check
        #any text to see if it might be such a case, and then backtrack
        #one place in the output.
        if self._justAfterEntity:
            match = leftover_entity.match(data)
            if match is not None:
                entityText = data[match.start():match.end()]
                #print 'entityText is "%s"' % entityText
                data = data[match.end():]   #continue with the latter text later
                #print 'leftover text is "%s"' % data
                
                if self.buf[-1] == '&amp;':   #we just put this there, we hope
                    self.buf.pop()
                self.buf.append('&%s' % entityText)
                
            self._justAfterEntity = False
        
        
        #if for any reason &, <, > is here, escape it.        
        data = escape(data)

        #are we in the right mode?
        if self._context == 'block':
            #naked text is naughty, cover it up!
            self.writeStartTag(self.DEFAULT_BLOCK_TAG)

        self.writeData(data)

    def handle_charref(self, name):
        if self._context == 'block':
            self.writeStartTag(self.DEFAULT_BLOCK_TAG)

        self.buf.append("&#%s;" % name)

    def handle_entityref(self, name):
        #print 'handle_entityref "%s"' % name
        if self._context == 'block':
            self.writeStartTag(self.DEFAULT_BLOCK_TAG)
        #copied from soup - what's it for?
        if not self.entitydefs.has_key(name):
            raise ValueError("Unknown entity '%s'" % name)
        self.buf.append("&%s;" % name)
        self._justAfterEntity = True

    def handle_comment(self, name):
        if not self.stripComments:
            self.buf.append('<!--' + name + '-->')
        
    def convert_entityref(self, name):
        if self.convertEntities:
            return sgmllib.SGMLParser.convert_entityref(self, name)
        else:
            return
        
    def unknown_starttag(self, tag, attrs):
        """ Delete all tags except for legal ones, and strip some obvious javascript

        The 'state machine' choices are all here, and in unknown_endtag. At any point,
        we know the current context, and the next tag which has its own context. If
        the next tag is expected, fine.  If not, we need to handle each state
        transition intelligently.

        """
        #print "unknown_starttag('%s'), buffer is '%s'" % (tag, ''.join(self.buf))
        tag = tag.lower()
        if tag in self.valid_tags:
            context = self._context    
            if tag in self.tagsAllowedInContext(self._context):
                #expected, write it out.  The writer will filter out any unexpected attributes.
                self.writeStartTag(tag, attrs)
            else:
                #Unexpected.  Each context combination has its own rules.
                #We have 6 contexts so 36 combinations, but in most we
                #just want to ignore the tag.
                nextContext = self.getContext(tag)

                #print ("Got unexpected <%s> in '%s', stack is '%s'" % (tag, context, ''.join(self.buf)))

                if context == 'inline':
                    if nextContext == 'block':
                        self.closeCurrentBlock()
                        self.writeStartTag(tag, attrs)
                    else:
                        pass   #if we get a naked tr, td, li, we'll just ignore it.
                elif context == 'block':
                    if nextContext == 'inline':  #e.g. <i> at start of document
                        #print 'so far: %s.  tag: %s' % (''.join(self.buf), tag)
                        self.writeStartTag(self.DEFAULT_BLOCK_TAG)
                        self.writeStartTag(tag, attrs)
                    elif nextContext in ('table', 'tr', 'list'):
                        #very out-of-context tag, ignore it. e.g. <p><tr>
                        pass
                elif context == 'table':
                    #anything but a tr or td is disallowed
                    if nextContext == 'tr':  #i.e. tag is a td
                        #they forgot the tr, repair it
                        self.writeStartTag('tr', {})
                        self.writeStartTag(tag, attrs)
                    else:
                        pass
                elif context == 'tr':
                    #expected a td, anything higher up means we need to close
                    if nextContext == 'table':
                        #got a tr - close the current tr
                        self.writeEndTag('tr')
                        self.writeStartTag(tag, attrs)
                    elif nextContext == 'block':
                        #close the whole table
                        self.writeEndTag('tr')
                        self.writeEndTag('table')
                        self.writeStartTag(tag, attrs)
                    else:
                        pass
                elif context == 'td':
                    #brutal for now, no tags allowed here yet, but
                    #should be like inline.
                    if nextContext == 'table':
                        #got a tr - close the td
                        self.writeEndTag('td')
                        self.writeEndTag('tr')
                        self.writeStartTag(tag, attrs)
                    elif nextContext == 'block':
                        #close the whole table
                        self.writeEndTag('td')
                        self.writeEndTag('tr')
                        self.writeEndTag('table')
                        self.writeStartTag(tag, attrs)
                    else:
                        pass
##                elif context == 'list':  #tag is li
##                    if nextContext == 'list':  #got another li, close the last li
##                        self.writeEndTag('li')
##                        self.writeStartTag(tag, attrs)
##                    elif nextContext == 'block':
##                        self.closeCurrentBlock()
##                        self.writeStartTag(tag, attrs)
##                    else:
##                        pass
                elif context == 'li':  #tag is li
                    if nextContext == 'list':  #got another li, close the last li
                        self.writeEndTag('li')
                        self.writeStartTag(tag, attrs)
                    elif nextContext == 'block':
                        self.closeCurrentBlock()
                        self.writeStartTag(tag, attrs)
                    else:
                        pass
                else:
                    raise ValueError("unexpected context '%s'" % context)


    def unknown_endtag(self, tag):
        #print "unknown_endtag('%s'), buffer is '%s', stack is %s" % (tag, ''.join(self.buf), self.openTagStack)
        tag = tag.lower()
        if tag in self.valid_tags:
            pending = self.pendingTag()
            if tag == pending:
                #all is wonderful, close it
                self.writeEndTag(tag)
            else:
                #normally, we just ignore unexpected end tags.
                #the stack will be closed at the end.  However
                #if we get a </table>, </ul> etc, it may make
                #sense to close.
                if tag == 'tr':
                    if pending == 'td':
                        #close it
                        self.writeEndTag('td')
                        self.writeEndTag(tag)
                elif tag in ('ul','ol'):
                    if pending == 'li':
                        self.writeEndTag('li')
                        self.writeEndTag(tag)
                elif tag == 'table':
                    if pending == 'td':
                        self.writeEndTag('td')
                        self.writeEndTag('tr')
                        self.writeEndTag(tag)
                    elif pending == 'tr':
                        self.writeEndTag('tr')
                        self.writeEndTag(tag)



        else:
            self.fixes.append("Ignoring unexpected end tag %s" % tag)

    def cleanupClosingTags(self):
        """ Append any missing closing tags. Called at end."""
        while self.openTagStack:
            tag = self.openTagStack.pop()
            self.buf.append("</%s>" % tag)
            self.fixes.append("appended missing end tag </%s>" % tag)

    def do_br(self, attrs):
        """Handles the <br/> tag.
        
        Called directly by sgmllib instead of unknown_starttag,
        because it's a singleton tag.  What we do depends on whether
        (a) breaks are allowed in this normalisation, and
        (b) we are inside a block or between them.
        
        As presently implemented, with breaksAllowed=False...
           <p>one<br/>two</p>   ->  <p>one</p><p>two</p>
        
        ..and multiple <br> tags beyond the first will create
        extra empty paragraphs.
        """
        if self.breaksAllowed:
            if self._context == 'inline':
                self.buf.append("<br/>")
        else:
            if self.target == 'block':
                if self._context == 'inline':
                    self.closeCurrentBlock()
                elif self._context == 'block':
                    #give them an empty para for each extra <br> they added
                    self.writeStartTag(self.DEFAULT_BLOCK_TAG)
                    self.writeEndTag(self.DEFAULT_BLOCK_TAG)
            else:
                self.buf.append(' ')  #if they had two lines in the input,
                #at least a space should separate them in the output.


    def unknown_charref(self, ref):
        print 'unknown charref', ref


#encoding detection tests
_testCp1252 = 'copyright %s trademark %s registered %s ReportLab! Ol%s!' % (chr(169), chr(153),chr(174), chr(0xe9))
_testUni = unicode(_testCp1252, 'cp1252')
_testUTF8 = _testUni.encode('utf-8')


TEST_CASES = [
    #list of (testname, mode, options, input, output) quads
    ('empty string', 'inline', None, "", ""),
    ('plain text not modified', 'inline', None, "Hello world", "Hello world"),
    ('simple ampersand preserved', 'inline', None, "Smith &amp; Jones", "Smith &amp; Jones"),

    #encoding detection
    ('utf8 OK', 'inline', None, _testUTF8, _testUTF8),
    ('cp1252 converted to utf8', 'inline', None, _testCp1252, _testUTF8),


    ('naked ampersand fixed', 'inline', None, "Smith & Jones", "Smith &amp; Jones"),
    ('respects trendy quotes', 'inline', None, "the &lt;&lt;ultra&gt;&gt; bar", "the &lt;&lt;ultra&gt;&gt; bar"),

    #tag processing
    ('respects tags and expected attributes', 'inline', None,
         'Click <a href="/help" alt="x">here</a> for help', 'Click <a href="/help">here</a> for help'),


    ('singleton <br> fixed', 'inline', None, 'line<br>two', 'line<br/>two'),
    ('cases fixed', 'inline', None, '<I>foo</i>', '<i>foo</i>'),

    #basic nesting repairs - pretty dumb algorithm, it ignores any unexpected end tags and closes any leftovers at the end.
    # a smarter parser might close on any 'end block' event.
    ('nesting test 1', 'inline', None, '<b>foo</i>', '<b>foo</b>'),
    ('nesting test 2', 'inline', None, '<b><i>foo</i>', '<b><i>foo</i></b>'),
    ('nesting test 3', 'block', None, '<p><b>foo', '<p><b>foo</b></p>'),

    #input cp1252 and smart quotes, see if it guesses - to be completed.



    #inline converted to block if not started off.
    ('block to inline', 'inline', None, '<p>foo</p>', 'foo'),
    ('inline to block', 'block', None, 'foo', '<p>foo</p>'),
    ('block respected', 'block', None, '<h1>foo</h1>', '<h1>foo</h1>'),
    ('blocks finished at end of document', 'block', None, '<h1>foo', '<h1>foo</h1>'),
    ('block finished off with correct tag', 'block', None, '<h1>foo<p>text</p>', '<h1>foo</h1><p>text</p>'),


    #detect and wrap text between / outside blocks
    ('naked text 1', 'block', None, 'first<p>second</p>', '<p>first</p><p>second</p>'),
    ('naked text 2', 'block', None, 'first<p>second</p>three', '<p>first</p><p>second</p><p>three</p>'),


    #br tag handling.  May or may not be allowed depending on mode.
    ('breaks rewritten with trailing slash', 'inline', None, 'one<br>two<br>three', 'one<br/>two<br/>three'),
    ('fixes expanded break', 'inline', None, 'one<br>two</br>three', 'one<br/>twothree'),

    ('breaks become paras in block', 'block', {'breaksAllowed':False}, 'one<br>two<br>three', '<p>one</p><p>two</p><p>three</p>'),
    ('breaks vanish inline', 'inline', {'breaksAllowed':False}, 'one<br>two<br>three', 'one two three'),
    ('breaks inside p preserved', 'block', None, '<p>Apple<br/><br/>Banana</p>', '<p>Apple<br/><br/>Banana</p>'),
    ('breaks with space inside p preserved', 'block', None, '<p>Apple<br /><br />Banana</p>', '<p>Apple<br/><br/>Banana</p>'),

    #tables
    ('tables start a block', 'block', {'breaksAllowed':False},
        'hello<br>line two<table></table>stuff after',
        '<p>hello</p><p>line two</p><table></table><p>stuff after</p>'),

    ('tables with tr and td', 'block', {'breaksAllowed':False},
        'hello<br>line two<table><tr><td>Hello</td></tr><tr><td>Dick & jane</td></tr></table>stuff after',
        '<p>hello</p><p>line two</p><table><tr><td>Hello</td></tr><tr><td>Dick &amp; jane</td></tr></table><p>stuff after</p>'),

    ('unfinished table gets closed', 'block', None, '<table><tr><td>Stuff', '<table><tr><td>Stuff</td></tr></table>'),
    ('unfinished table gets closed 2', 'block', None, '<table><tr><td>Stuff<p>more</p>', '<table><tr><td>Stuff</td></tr></table><p>more</p>'),

    ('tables with tr and td bad block', 'block', {'breaksAllowed':False},
        'hello<br>line two<table><tr><td>Hello</td><tr><td>Dick & jane</table>stuff after',
        '<p>hello</p><p>line two</p><table><tr><td>Hello</td></tr><tr><td>Dick &amp; jane</td></tr></table><p>stuff after</p>'),

    #lists
    ('list copied OK','block',None, '<ul><li>one</li><li>two</li></ul>','<ul><li>one</li><li>two</li></ul>'),
    ('li tags closed early','block',None, '<ul><li>one<li>two</ul>','<ul><li>one</li><li>two</li></ul>'),
    ('li tags closed early 2','block',None, '<ul><li>one<li>two</ul>end','<ul><li>one</li><li>two</li></ul><p>end</p>'),

    ('Bad list html','block',None, '<li>one</li><ul><li>two</li></ul>','<p>one</p><ul><li>two</li></ul>'),
    ('Bad table html','block',None, '<td>one</td><table><tr>two</tr></table>','<p>one</p><table><tr>two</tr></table>'),

    ('Html with unicode characters', 'block', None, """  <p>Refuge luxueux et original située à 50 km à l’est de Mahé, Frégate Island Private se compose de 16 villas privées.  La règle qui prévaut dans l’île est la préservation et la protection d’un environnement naturel unique.  Dans ce lieu harmonieusement préservé, l’hôtel propose un somptueux décor et des installations luxueuses qui vous permettront de vivre à l’écart du monde et de jouir d’un repos complet dans une totale intimité.  La beauté naturelle de l’île, une cuisine gastronomique délicieuse, un large éventail d’activités sportives et de loisirs, deux piscines (dont une de 25 mètres pour ceux qui souhaitent faire des longueurs!): tout est organisé pour répondre à vos attentes les plus exigeantes.  Sont également accessibles le “Castaway Kids Club” et “The Rock Spa”.</p>""", """<p>Refuge luxueux et original située à 50 km à l’est de Mahé, Frégate Island Private se compose de 16 villas privées.  La règle qui prévaut dans l’île est la préservation et la protection d’un environnement naturel unique.  Dans ce lieu harmonieusement préservé, l’hôtel propose un somptueux décor et des installations luxueuses qui vous permettront de vivre à l’écart du monde et de jouir d’un repos complet dans une totale intimité.  La beauté naturelle de l’île, une cuisine gastronomique délicieuse, un large éventail d’activités sportives et de loisirs, deux piscines (dont une de 25 mètres pour ceux qui souhaitent faire des longueurs!): tout est organisé pour répondre à vos attentes les plus exigeantes.  Sont également accessibles le “Castaway Kids Club” et “The Rock Spa”.</p>"""),
    ('Para with anchor tags', 'block', None, '<p><a href="http://www.yahoo.com">Yahoo</a></p>', '<p><a href="http://www.yahoo.com">Yahoo</a></p>'),


    #Entity
    ('Entities remain untouched', 'block', None, '<p>&copy;</p>', '<p>&copy;</p>'),
    ('Entities remain untouched big', 'block', None,
    """<p><img src="dick &copy; jane" alt=""/></p>""",
    """<p><img src="dick &copy; jane" alt=""/></p>"""
    ),
    ('Entities can be converted on demand', 'block', {'convertEntities': True},
    """<p><img src="dick &copy; jane" alt=""/></p>""",
    """<p><img src="dick © jane" alt=""/></p>"""
    ), 
    ('Invalid entities', 'block', None, '<p>Apple&sup2;</p>', '<p>Apple&sup2;</p>'),


    #images
    ('image inside block', 'block', None, '<img src="foo.jpg"/>', '<p><img src="foo.jpg" alt=""/></p>'),
    ('image starts new block', 'block', None, 'before<img src="foo.jpg"/>after', '<p>before<img src="foo.jpg" alt=""/>after</p>'),
    ('alt attribute gets added to img', 'block', None, '<p><img src="apple.jpg"/></p>', '<p><img src="apple.jpg" alt=""/></p>'),
    ('old alt attribute will not be touched', 'block', None, '<p><img src="apple.jpg" alt="happy"/></p>', '<p><img src="apple.jpg" alt="happy"/></p>'),
    ('html with images', 'block', None,
    """Go& There can be yacht.<br /><br />Seychelles n<br /><strong>Sunday:</strong><br />We proposMah&eacute;, principal iort.<img height="225" alt="Sailing in Seychelles - Copyright Koos van der Lende" hspace="5" width="150" align="right" vspace="5" src="/media/Sailbad.jpg" /><br /><br />This imposing, verdant granite island whose topmost peak towers to over 900m boasts over 65 beaches <p><img height="134" alt="Sailing in Seychelles - Copyright Sunsail / Jonathan Smith" hspace="5" width="200" align="right" vspace="5" src="/media/Sunsail_of_Sisters_Island.jpg" /></p><br /></strong>It will not be lon""",
    """<p>Go&amp; There can be yacht.<br/><br/>Seychelles n<br/>Sunday:<br/>We proposMah&eacute;, principal iort.<img height="225" alt="Sailing in Seychelles - Copyright Koos van der Lende" width="150" src="/media/Sailbad.jpg"/><br/><br/>This imposing, verdant granite island whose topmost peak towers to over 900m boasts over 65 beaches </p><p><img height="134" alt="Sailing in Seychelles - Copyright Sunsail / Jonathan Smith" width="200" src="/media/Sunsail_of_Sisters_Island.jpg"/></p><p>It will not be lon</p>"""
    ),
    ('comments would be preserved', 'block', None,
    """<p><!-- page_break --></p>""",
    """<p><!-- page_break --></p>"""
    ),
    ('paras with break', 'block', None,
    """sajdfdskjfhsk <br/> <br/>asfdfkshkfjsjfks""",
    """<p>sajdfdskjfhsk <br/> <br/>asfdfkshkfjsjfks</p>"""
    ),
    ('Complex para with breaks', 'block', None,
    """<p>Il y a trois langues officielles aux Seychelles: le Cr&#233;ole (patois bas&#233; sur le Fran&#231;ais), l&#8217;Anglais et le Fran&#231;ais. &#160;Beaucoup de seychellois parlent &#233;galement couramment l&#8217;Italien et l&#8217;Allemand.<br /><br />Voici quelques expressions cr&#233;oles bien utiles:</p><table class="btxt" cellspacing="2" cellpadding="2" align="left" border="0"><tbody><tr class="tablerow"><td class="tableheader" width="160">Fran&#231;ais</td><td class="tableheader" width="160">Creole</td></tr><tr bgcolor="#eeeeee"><td width="160">Bonjour</td><td width="160">Bonzour</td></tr><tr bgcolor="#eeeeee"><td width="160">Au revoir</td><td width="160">Orevwar</td></tr><tr bgcolor="#eeeeee"><td width="160">Comment allez vous?</td><td width="160">Ki dir?</td></tr><tr bgcolor="#eeeeee"><td width="160">Merci</td><td width="160">Mersi</td></tr><tr bgcolor="#eeeeee"><td width="160">O&#249;?</td><td width="160">Kote?</td></tr><tr bgcolor="#eeeeee"><td width="160">S&#8217;il vous pla&#206;t</td><td width="160">Silvouple</td></tr><tr bgcolor="#eeeeee"><td width="160">Non</td><td width="160">Non</td></tr><tr bgcolor="#eeeeee"><td width="160">Oui</td><td width="160">Wi</td></tr><tr bgcolor="#eeeeee"><td width="160">Je ne comprends pas</td><td width="160">Mon pa konpran</td></tr><tr bgcolor="#eeeeee"><td width="160">&#231;a me pla&#206;t</td><td width="160">Mon kontan</td></tr><tr bgcolor="#eeeeee"><td width="160">Comment &#231;a va?</td><td width="160">Konman sava?</td></tr><tr bgcolor="#eeeeee"><td width="160">Qu&#8217;est ce que c&#8217;est?</td><td width="160">Kisisa?</td></tr></tbody></table>""",
    """<p>Il y a trois langues officielles aux Seychelles: le Cr&#233;ole (patois bas&#233; sur le Fran&#231;ais), l&#8217;Anglais et le Fran&#231;ais. &#160;Beaucoup de seychellois parlent &#233;galement couramment l&#8217;Italien et l&#8217;Allemand.<br/><br/>Voici quelques expressions cr&#233;oles bien utiles:</p><table><tr><td>Fran&#231;ais</td><td>Creole</td></tr><tr><td>Bonjour</td><td>Bonzour</td></tr><tr><td>Au revoir</td><td>Orevwar</td></tr><tr><td>Comment allez vous?</td><td>Ki dir?</td></tr><tr><td>Merci</td><td>Mersi</td></tr><tr><td>O&#249;?</td><td>Kote?</td></tr><tr><td>S&#8217;il vous pla&#206;t</td><td>Silvouple</td></tr><tr><td>Non</td><td>Non</td></tr><tr><td>Oui</td><td>Wi</td></tr><tr><td>Je ne comprends pas</td><td>Mon pa konpran</td></tr><tr><td>&#231;a me pla&#206;t</td><td>Mon kontan</td></tr><tr><td>Comment &#231;a va?</td><td>Konman sava?</td></tr><tr><td>Qu&#8217;est ce que c&#8217;est?</td><td>Kisisa?</td></tr></table>"""
    ),
    ('Para with one break gets split when breaks not allowed', 'block', {'breaksAllowed' : False},
    """<p>Part one<br />Part two</p>""",
    """<p>Part one</p><p>Part two</p>"""
    ),    

    ('Para with double <br/> when breaks not allowed', 'block', {'breaksAllowed' : False},
    """<p>Part one<br /><br />Part two</p>""",
    """<p>Part one</p><p></p><p>Part two</p>"""
    ),    
    ('Para with treble <br/> when breaks not allowed', 'block', {'breaksAllowed' : False},
    """<p>Part one<br /><br /><br />Part two</p>""",
    """<p>Part one</p><p></p><p></p><p>Part two</p>"""
    ),

    ('Entity examples', 'inline', None,
    '&amp;', '&amp;'
    ),
    ('Entity examples1', 'inline', None,
    '&eacute;', '&eacute;'
    ),
    ('Fix doubled entity', 'inline', None,
    '&amp;eacute; and so on', '&eacute; and so on'
    ),
    ('Preserves numeric entity', 'inline', None,
    '&#233;', '&#233;'
    ),
    ('Fix doubled numeric entity', 'inline', None,
    '&amp;#233;', '&#233;'
    ),
    ('Fix tribled numeric entity', 'inline', None,
    '&amp;amp;#233;', '&amp;#233;'
    ),
    ('div  gets converted to para', 'block', None,
    """<div id="hello">fashd kjah asfh aasfda</div>""",
    """<p>fashd kjah asfh aasfda</p>"""
    ),
    ]

def test(verbose=False):
    for (testName, mode, options, input, expected) in TEST_CASES:
        #options may be a dict
        if verbose:
            print
        print testName,'...',
        if options is None:
            options = {}

        c = Cleaner(target=mode, **options)
        out = c.process(input)
        if out == expected:
            print 'OK'
        else:
            print 'failed.  \n  Input\n    %s\n  Expected\n    %s\n  Got\n    %s' % (input, expected, out)
        if c.fixes:
            if verbose:
                print '  fixes applied:'
                for fix in c.fixes:
                    print '    %s' % fix


if __name__=='__main__':

    import sys
    args = sys.argv[:]
    verbose = ('-v' in args) or ('--verbose' in args)
    test(verbose=verbose)


    #import doctest, html_cleaner
    #doctest.testmod(html_cleaner)
